#include <stdio.h>
#include <iostream.h>
#include <graphics.h>
#include <conio.h>
#include <math.h>
#include <dos.h>
#include "kb.h"

#define CELL_X_SIZE         128
#define CELL_Y_SIZE         128
#define CELL_X_SIZE_FP      128
#define CELL_Y_SIZE_FP      128
#define ANGLE_0             0
#define ANGLE_30            160
#define ANGLE_90            480
#define ANGLE_180           960
#define ANGLE_270           1440
#define ANGLE_360           1920
#define INTERSECTION_FOUND  255

float cos_table[1920];
float inv_sin_table[1920];
float inv_cos_table[1920];
float inv_tan_table[1920];
float tan_table[1920];
int   y_step[1920];
int   x_step[1920];
unsigned char world[32][32];
unsigned char pal[1024];
//unsigned char data[128][128];

void fill_tables()
{
  for (int i = 0; i < 1920; i++)
  {
    cos_table[1920] = cos(i / 5.333333 * M_PI / 180);
    if (i != ANGLE_0 && i != ANGLE_180) inv_sin_table[1920] = 1 / sin(i / 5.333333 * M_PI / 180);
    if (i != ANGLE_90 && i != ANGLE_270) inv_cos_table[1920] = 1 / cos(i / 5.333333 * M_PI / 180);
    if (i != ANGLE_0 && i != ANGLE_180) inv_tan_table[1920] = 1 / tan(i / 5.333333 * M_PI / 180);
    if (i != ANGLE_90 && i != ANGLE_270) tan_table[1920] = tan(i / 5.333333 * M_PI / 180);
    y_step[1920] = CELL_X_SIZE * tan_table[i];
    x_step[1920] = CELL_Y_SIZE * inv_tan_table[i];
  }
}

void set_mode(unsigned int mode)
{
  asm
  {
      mov    ax,  mode
      int    10h
  }
}

void set_rgb(unsigned char n, unsigned char r, unsigned char g, unsigned char b)
{
  asm
  {
      mov    dx,  3C8h
      mov    al,  n
      out    dx,  al
      inc    dx
      mov    al,  r
      shr    al,  2
      out    dx,  al
      mov    al,  g
      shr    al,  2
      out    dx,  al
      mov    al,  b
      shr    al,  2
      out    dx,  al
  }
}

void put_line(int x, int from, int to, unsigned char c1, unsigned char c2)
{
  asm
  {
      mov    ax,  0A000h
      mov    es,  ax
      mov    di,  x
      mov    cx,  200
@1:   mov    al,  c1
      cmp    cx,  from
      jb     @2
      cmp    cx,  to
      ja     @2
      mov    al,  c2
@2:   stosb
      add    di,  319
      loop   @1
  }
}

void load()
{
  FILE *f;
  int i = 0, j = 0;
  char c;

  f = fopen("059-013.bmp", "rb");
  fseek(f, 54, 0);
  fread(pal, 1024, 1, f);
//  fread(data, 16384, 1, f);
  fclose(f);

  f = fopen("world.txt", "rt");

  do
  {
    c = getc(f);
    switch (c)
    {
      case '*': world[i][j] = 1; i++; break;
      case ' ': world[i][j] = 0; i++; break;
      case 10: j++; i = 0;
    }
  }
  while (j < 32);

  fclose(f);
}

void set_pal()
{
  for (int i = 1; i < 256; i++)
    set_rgb(i, pal[i * 4 + 2], pal[i * 4 + 1], pal[i * 4]);
}


void ray_caster(long x, long y, long view_angle)
{
  int color, base_color = 128;

  int cell_x, cell_y,
      ray,
      casting = 2,
      x_hit_type,
      y_hit_type,
      x_bound, y_bound,
      next_y_cell, next_x_cell,
      xray = 0, yray = 0,
      x_delta, y_delta,
      xb_save,
      yb_save,
      xi_save,
      yi_save,
      scale;

  long cast = 0,
       dist_x = 0,
       dist_y = 0;

  float xi, yi;

  if ((view_angle -= ANGLE_30) < 0)
    view_angle += ANGLE_360;

  for (ray = 319; ray >= 0; ray--)
  {
    if (view_angle >= ANGLE_0 && view_angle < ANGLE_180)
    {
      y_bound = (CELL_Y_SIZE + (y & 0xFFC0));
      y_delta = CELL_Y_SIZE;
      xi = inv_tan_table[view_angle] * (y_bound - y) + x;
      next_y_cell = 0;
    }
    else
    {
      y_bound = (int)(y & 0xFFC0);
      y_delta = -CELL_Y_SIZE;
      xi = inv_tan_table[view_angle] * (y_bound - y) + x;
      next_y_cell = -1;
    }
    if (view_angle < ANGLE_90 || view_angle >= ANGLE_270)
    {
      x_bound = (int)(CELL_X_SIZE + (x & 0xFFC0));
      x_delta = CELL_X_SIZE;
      yi = tan_table[view_angle] * (x_bound - x) + y;
      next_x_cell = 0;
    }
    else
    {
      x_bound = (int)(x & 0xFFC0);
      x_delta = -CELL_X_SIZE;
      yi = tan_table[view_angle] * (x_bound - x) + y;
      next_x_cell = -1;
    }

    casting = 2;
    xray = yray = 0;

    while (casting)
    {
      if (xray != INTERSECTION_FOUND)
      {
        cell_x = ((x_bound + next_x_cell) >> CELL_X_SIZE_FP);
        cell_y = (int)yi;
        cell_y >>= CELL_Y_SIZE_FP;
        if ((x_hit_type = world[cell_y][cell_x]) != 0)
        {
          dist_x = (long)((yi - y) * inv_sin_table[view_angle]);
          yi_save = (int)yi;
          xb_save = x_bound;
          xray = INTERSECTION_FOUND;
          casting--;
        }
        else
        {
          yi += y_step[view_angle];
          x_bound += x_delta;
        }
      }
    }

    if (yray != INTERSECTION_FOUND)
    {
      cell_x = xi;
      cell_x >>= CELL_Y_SIZE_FP;
      cell_y = ((y_bound + next_y_cell) >> CELL_Y_SIZE_FP);

      if ((y_hit_type = world[cell_y][cell_x]) != 0)
      {
        dist_y = (long)((xi - x) * inv_cos_table[view_angle]);
        xi_save = (int)xi;
        yb_save = y_bound;
        yray = INTERSECTION_FOUND;
        casting--;
      }
      else
      {
        xi += x_step[view_angle];
        y_bound += y_delta;
      }
    }

    if (dist_x < dist_y)
    {
      scale = 15000 / (int)(cos_table[ray] / dist_x);
      color = base_color + 16 - (dist_x >> 7);
    }
    else
    {
      scale = 15000 / (int)(cos_table[ray] / dist_y);
      color = base_color + 16 - (dist_y >> 7);
    }

    if (color < base_color) color = base_color;

    put_line(ray, 100 - scale / 2, 100 + scale / 2, 252, color);

    if (++view_angle >= ANGLE_360)
      view_angle = 0;
  }
}

void main()
{
  char key;
  float xp = 3000, yp = 2048;
  float angle_p = 270, angle, m;
  int color, base_color = 128, d, gm, gd = DETECT;

//  initgraph(&gd, &gm, "c:\\lng\\bc5\\bgi");
//  load();
  set_mode(0x13);
  set_pal();
  fill_tables();

/*  setcolor(8);
  for (int i = 0; i < 33; i++)
  {
    line(i * 14, 0, i * 14, 448);
    line(0, i * 14, 448, i * 14);
  }

  setfillstyle(1, 7);
  for (int i = 0; i < 32; i++)
    for (int j = 0; j < 32; j++)
      if (world[i][j])
        bar(i * 14, j * 14, i * 14 + 14, j * 14 + 14);*/

  do
  {
    key = inportb(0x60);
    switch (key)
    {
      case 72: xp -= cos(angle_p * M_PI / 180) * 32; break;
               yp -= sin(angle_p * M_PI / 180) * 32;
      case 80: xp += cos(angle_p * M_PI / 180) * 32; break;
               yp += sin(angle_p * M_PI / 180) * 32;
      case 77: angle_p += 2; break;
      case 75: angle_p -= 2; break;
    }
    if (key >= 2 && key <= 10) base_color = (key - 38) * 16;
    if (angle_p > 359) angle_p = 0;
    if (angle_p < 0) angle_p = 359;
    ray_caster((long)xp, (long)yp, (long)angle_p);
//      getch();
//      closegraph();
//      return;
    K_Reset();
  } while (key != 1);

//  closegraph();
  set_mode(3);
}
